#ifndef __USE_GNU
#define __USE_GNU
#endif
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>

#include <pthread.h>

#include <math.h>

#include <sched.h>

#include <errno.h>

#include "cachetest.h"

int test_z;
int test_z_float;

int loops;
double loops_float;

int num_cores;

#define PTHREAD_AFFINITY

int stick_this_thread_to_core(int core_id) {

   int res;
   (void)res;

   if ( core_id == -1 ) return 0;

   if (core_id < 0 || core_id >= num_cores) {
     printf("ERROR: %d\n",core_id);
     return EINVAL;
   }

   /* S E T */

   cpu_set_t cpuset;

   CPU_ZERO(&cpuset);
   CPU_SET(core_id, &cpuset);

#ifdef  PTHREAD_AFFINITY
   pthread_t current_thread = pthread_self();

   res = pthread_setaffinity_np(current_thread, sizeof(cpu_set_t), &cpuset);
#else
  sched_setaffinity(0, sizeof(cpuset), &cpuset);
#endif

#ifdef TEST_AFFINITY
   printf("Res %d\n", res);

   /* T E S T */

   cpu_set_t cpuset_test;

   CPU_ZERO(&cpuset_test);
   res = pthread_getaffinity_np(current_thread, sizeof(cpu_set_t), &cpuset_test);

   printf("Test-res %d\n", res);

   printf("Want %d -> count: %d set %d: %d\n", core_id, CPU_COUNT(&cpuset_test), core_id, CPU_ISSET(core_id, &cpuset_test));
#endif

   return 0;
}

int biggest_prim(int z) {
  int biggest_prim = 2;
  int i,t;

  for(i = 3; i <= z; i++) {

    int is_prim = 1;
    for(t = 2; t < i; t++) {
      if ( i%t == 0 ) is_prim = 0;
    }

    if ( is_prim ) {
      if ( i > biggest_prim ) biggest_prim = i;
    }
  }

  return biggest_prim;
}

float biggest_prim_float(float z) {
  float biggest_prim = 2;
  float i,t;
  float is_prim;

  for( i = 3.0; i <= z; i = i + 1.0) {

    is_prim = 1.0;
    for(t = 2.0; t < i; t = t  + 1.0) {
      if ( fmod(i,t) == 0 ) is_prim = 0.0;
    }

    if ( is_prim == 0.0 ) {
      if ( i > biggest_prim ) biggest_prim = i;
    }
  }

  return biggest_prim;
}

void test_thread(void *argv)
{
  int i;
  int cpu_core = (int)argv;

  stick_this_thread_to_core(cpu_core);

  for(i = 0; i < loops; i++) {
    biggest_prim(test_z);
  }
}

void test_thread_float(void *argv)
{
  float i;
  int cpu_core = (int)argv;

  stick_this_thread_to_core(cpu_core);

  for(i = 0; i < loops_float; i = i + 1.0) {
    biggest_prim_float((double)test_z_float);
  }
}

void test_mat_thread(void *argv)
{
  //int i;
  int cpu_core = (int)argv;

  stick_this_thread_to_core(cpu_core);

  //for(i = 0; i < loops; i++) {
    run_mat_test();
  //}
}

int main(int argc, char **argv)
{
  int no_threads = 0;
  int mode = 0;
  int task_mode = 0;

  num_cores = sysconf(_SC_NPROCESSORS_ONLN);

  if ( argc < 5 ) {
    printf("Usage:\n %s #no_threads #no_loops mode task_mode!\n",argv[0]);
    printf("\nMode:\n");
    printf(" 0: thread 0 & 1 -> cpu 0 ; thread 2 & 3 -> cpu 1 (cpu = i/2)\n");
    printf(" 1: thread 0 & 2 -> cpu 0 ; thread 1 & 3 -> cpu 1 (cpu = i%%2)\n");
    printf(" 2: thread 0 -> cpu 0 ; thread 1 -> cpu 1 ; ...   (cpu = i)\n");
    printf(" 3: thread 0 -> cpu 0 ; thread 1 -> cpu 2 ; thread 2 -> cpu 1 ; thread 3 -> cpu 3 (cpu = (i%%2)*2)\n");
    printf(" 4: thread 0 -> cpu ? ; thread 1 -> cpu ? ; thread 2 -> cpu ? ; thread 3 -> cpu ? (cpu = linux)\n\n");

    printf("Task-Mode:\n");
    printf(" 0: thread 0-3 -> task 0 (task = 0)\n");
    printf(" 1: thread 0-3 -> task 1 (task = 1)\n");
    printf(" 2: thread 0 & 2 -> task 0 ; thread 1 & 3 -> task 1 (task = i%%2)\n");
    printf(" 3: thread 0 & 1 -> task 0 ; thread 2 & 3 -> task 1 (task = i/2)\n");
    printf(" 4: task = 2 -> mat\n");
    printf(" 5: thread 0 & 2 -> task 0 ; thread 1 & 3 -> task 2\n");

    exit(0);
  }

  no_threads = atoi(argv[1]);
  loops = atoi(argv[2]);
  loops_float = (float)loops;
  mode = atoi(argv[3]);
  task_mode = atoi(argv[4]);


  test_z = 9000;
  test_z_float = 3920;

  printf("CPU-Cores: %d\n\nStart %d no_threads and %d no_loops with mode %d ...\n\n", num_cores, no_threads, loops, mode);

  pthread_t *threads = (pthread_t *)malloc((no_threads) * sizeof(pthread_t));

  int i = 0;
  int core_id;
  int task;

  for ( ; i < no_threads; i++ ) {
    switch (mode) {
      case 0:
        /*
         * cpu = i/2: thread 0 & 1 -> cpu 0 ; thread 2 & 3 -> cpu 1
         */
        core_id = (i/2);
        break;
      case 1:
        /*
         * cpu = i%2: thread 0 & 2 -> cpu 0 ; thread 1 & 3 -> cpu 1
         */
        core_id = (i%2);
        break;
      case 2:
        /*
         * cpu = i: thread 0 -> cpu 0 ; thread 1 -> cpu 1 ; ...
         */
        core_id = i;
        break;
      case 3:
        /*
         * cpu = (i%2)*2 + (i/2): thread 0 -> cpu 0 ; thread 1 -> cpu 2 ; thread 2 -> cpu 1 ; thread 3 -> cpu 3
         */
        core_id = (i%2)*2 + (i/2);
        break;
      case 4:
        /*
         * cpu = linux
         */
        core_id = -1;
        break;
    }

    switch (task_mode) {
      case 0:
        /*
         * task = 0: thread 0-3 -> task 0
         */
        task = 0;
        break;
      case 1:
        /*
         * task = 1: thread 0-3 -> task 1
         */
        task = 1;
        break;
      case 2:
        /*
         * task = i%2: thread 0 & 2 -> task 0 ; thread 1 & 3 -> task 1
         */
        task = i%2;
        break;
      case 3:
        /*
         * task = i/2: thread 0 & 1 -> task 0 ; thread 2 & 3 -> task 1
         */
        task = i/2;
        break;
      case 4:
        /*
         * task = 2 -> mat
         */
        task = 2;
        break;
      case 5:
        /*
         * task = i%2: thread 0 & 2 -> task 0 ; thread 1 & 3 -> task 2
         */
        task = (i%2) * 2;
        break;
    }

    printf("Thread %d (Task %d) on CPU-Core %d\n", i, task, core_id);

    switch (task) {
      case 0:
        pthread_create(&threads[i],(pthread_attr_t*)NULL,(void *)&test_thread,(void*)core_id);
        break;
     case 1:
        pthread_create(&threads[i],(pthread_attr_t*)NULL,(void *)&test_thread_float,(void*)core_id);
        break;
     case 2:
        pthread_create(&threads[i],(pthread_attr_t*)NULL,(void *)&test_mat_thread,(void*)core_id);
        break;
   }

  }

  for ( i = 0; i < no_threads; i++ ) {
    pthread_join(threads[i],NULL);
  }

  return 0;
}
